// Decompile selected function recursively
// @category _Reman3
// @menupath Reman3.Dump Current Function (recursive)
// @importpackage org.sqlite

import java.util.ArrayList;
import java.util.List;

import ghidra.app.script.GhidraScript;
import ghidra.program.model.listing.Function;
import re3lib.FunctionDatabase;
import re3lib.FunctionDumper;
import re3lib.GlobalDumper;
import re3lib.PCallTracer;
import re3lib.RemanConfig;
import re3lib.TypeDumper;

public class DumpCurrentFunctionRecursive extends GhidraScript {
  @Override
  public void run() throws Exception {
    RemanConfig.INSTANCE = new RemanConfig(this);
    RemanConfig.INSTANCE.createDirectories();

    try (FunctionDatabase functionDatabase = new FunctionDatabase(this)) {
      GlobalDumper globalDumper = new GlobalDumper(this, functionDatabase);

      FunctionDumper functionDumper = new FunctionDumper(this, functionDatabase, globalDumper);

      PCallTracer tracer = new PCallTracer();
      tracer.setBlacklist(functionDumper.functionAddrBlackList);
      tracer.traceCalls(getFunctionContaining(currentAddress));

      List<Function> functionsToDump = new ArrayList<>();
      List<Function> functionsToDumpNew = new ArrayList<>();
      for (Function func : tracer.out) {
        List<FunctionDatabase.FunctionEntry> entries = functionDatabase.findEntriesByAddress(func.getEntryPoint());
        boolean shouldDump = true;
        boolean isNew = true;
        for (FunctionDatabase.FunctionEntry entry : entries) {
          if (entry.type == FunctionDatabase.Type.Fix) {
            shouldDump = false;
          }
          if (entry.type == FunctionDatabase.Type.Fix || entry.type == FunctionDatabase.Type.Auto) {
            isNew = false;
          }
        }
        if (!shouldDump)
          continue;

        println("Dump: " + func.getName());
        functionsToDump.add(func);

        if (isNew)
          functionsToDumpNew.add(func);
      }

      if (!functionsToDump.isEmpty()) {
        String newOpt = "Only new (" + functionsToDumpNew.size() + ")";
        String okOpt = "Yes (" + functionsToDump.size() + ")";
        String choice = askChoice("Confirmation", "About to generate " + functionsToDump.size() + " functions ("
            + functionsToDumpNew.size() + " new), continue?",
            new ArrayList<String>() {
              {
                add(okOpt);
                add(newOpt);
                add("No");
              }
            }, okOpt);
        if (choice == okOpt) {
        } else if (choice == newOpt) {
          functionsToDump = functionsToDumpNew;
        } else {
          return;
        }

        for (Function func : functionsToDump) {
          functionDumper.dump(func);
        }

        if (functionDumper.createdFile)
          RemanConfig.INSTANCE.touchCMakeTimestamp();

        globalDumper.dumpGlobals();
      }

      // Dump types
      TypeDumper dumper = new TypeDumper(this);
      dumper.run();
    }
  }
}
